/*
 * Copyright 2006 Dmitry A.Steklenev
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *    3. The name of the author may not be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define  INCL_DOS
#define  INCL_ERRORS
#include <os2.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <nerrno.h>
#include <netdb.h>
#include <sys/socket.h>

#include <utilfct.h>
#include <decoder_plug.h>
#include <debuglog.h>

#include "xio.h"
#include "xio_protocol.h"
#include "xio_file.h"
#include "xio_ftp.h"
#include "xio_http.h"
#include "xio_cddb.h"
#include "xio_syncbuffer.h"
#include "xio_asyncbuffer.h"
#include "xio_socket.h"

#define XIO_SERIAL 0x41290837

static char   http_proxy_host[XIO_MAX_HOSTNAME];
static int    http_proxy_port;
static char   http_proxy_user[XIO_MAX_USERNAME];
static char   http_proxy_pass[XIO_MAX_PASSWORD];
static u_long http_proxy_addr;

static int  buffer_size    = 32768;
static int  buffer_wait    = 0;
static int  buffer_fill    = 30;
static int  socket_timeout = 30;

/* Serializes access to the library's global data. */
static Mutex mutex;


static void DLLENTRY xio_observer( const char* metabuff, long pos, long pos64, void* arg )
{ XFILE* x = (XFILE*)arg;
  DEBUGLOG(("xio:xio_observer(%s, %lx,%lx, %p) - %p, %p\n", metabuff, pos, pos64, arg, x->s_observer, x->s_metabuff));
  
  if( x->s_observer && x->s_metabuff )
  { strlcpy( x->s_metabuff, metabuff, x->s_metasize );
    WinPostMsg( x->s_observer, WM_METADATA, MPFROMP( x->s_metabuff ), MPFROMLONG( pos ));
  }
}

/* Open file. Returns a pointer to a file structure that can be used
   to access the open file. A NULL pointer return value indicates an
   error. */
XFILE* DLLENTRY
xio_fopen( const char* filename, const char* mode )
{
  DEBUGLOG(("xio_fopen(%s, %s)\n", filename, mode));
  XFILE* x;

  if( !filename || !*filename ) {
    errno = ENOENT;
    return NULL;
  }
  x = new XFILE();

  if( strnicmp( filename, "http:", 5 ) == 0 )
  {
    //x->scheme   = XIO_HTTP;
    x->protocol = new XIOhttp();
  }
  else if( strnicmp( filename, "ftp:", 4 ) == 0 )
  {
    if( *http_proxy_host ) {
      //x->scheme   = XIO_HTTP;
      x->protocol = new XIOhttp();
    } else {
      //x->scheme   = XIO_FTP;
      x->protocol = new XIOftp();
    }
  }
  else if( strnicmp( filename, "cddbp:", 6 ) == 0 )
  {
    //x->scheme   = XIO_CDDB;
    x->protocol = new XIOcddb();
  } else {
    //x->scheme   = XIO_FILE;
    x->protocol = new XIOfile();
  }

  if( !x->protocol ) {
    delete x;
    return NULL;
  }

  if( strchr( mode, 'r' )) {
    if( strchr( mode, '+' )) {
      x->oflags = XO_READ | XO_WRITE;
    } else {
      x->oflags = XO_READ;
    }
  } else if( strchr( mode, 'w' )) {
    if( strchr( mode, '+' )) {
      x->oflags = XO_WRITE | XO_CREATE | XO_TRUNCATE | XO_READ;
    } else {
      x->oflags = XO_WRITE | XO_CREATE | XO_TRUNCATE;
    }
  } else if( strchr( mode, 'a' )) {
    if( strchr( mode, '+' )) {
      x->oflags = XO_WRITE | XO_CREATE | XO_APPEND | XO_READ;
    } else {
      x->oflags = XO_WRITE | XO_CREATE | XO_APPEND;
    }
  } else {
    delete x;
    errno = EINVAL;
    return NULL;
  }
  
  if (strchr(mode, 'R'))
    x->oflags |= XO_NOBUFFER;
  if (strchr(mode, 'X'))
    x->oflags |= XO_ASYNCBUFFER;
  if (strchr(mode, 'U'))
    x->oflags |= XO_NOMUTEX;

  XSFLAGS support = x->protocol->supports();
  if((( x->oflags & XO_READ   ) &&
      ( x->oflags & XO_WRITE  ) && !( support & XS_CAN_READWRITE )) ||
     (( x->oflags & XO_READ   ) && !( support & XS_CAN_READ      )) ||
     (( x->oflags & XO_WRITE  ) && !( support & XS_CAN_WRITE     )) ||
     (( x->oflags & XO_CREATE ) && !( support & XS_CAN_CREATE    ))  )
  {
    delete x;
    errno = EINVAL;
    return NULL;
  }

  if( x->protocol->open( filename, x->oflags ) != 0 ) {
    delete x;
    return NULL;
  }

  // Set up buffer
  if ( !(x->oflags & XO_NOBUFFER) && !( x->oflags & XO_WRITE ) )
//    && !( x->protocol->supports() & XS_NOT_BUFFERIZE ) )
  { int size = xio_buffer_size();
    XIObuffer* buffer;
    if ((x->oflags & XO_ASYNCBUFFER) && size)
      buffer = (XIObuffer*)new XIOasyncbuffer( x->protocol, xio_buffer_size(), x->protocol->blocksize );
    else
      buffer = new XIOsyncbuffer( x->protocol, x->protocol->blocksize );
    //XIObuffer* buffer = new XIOsyncbuffer( x->protocol, x->protocol->blocksize );
    if (buffer->init())
      x->protocol = buffer;
    else
      delete buffer;
  }
 
  if (!(x->oflags & XO_NOMUTEX))
    x->mtx = new Mutex();
  
  // We got it!  
  x->serial = XIO_SERIAL;

  x->protocol->set_observer( xio_observer, x );
  DEBUGLOG(("xio_open: %p\n", x));
  return x;
}

/* Reads up to count items of size length from the input file and
   stores them in the given buffer. The position in the file increases
   by the number of bytes read. Returns the number of full items
   successfully read, which can be less than count if an error occurs
   or if the end-of-file is met before reaching count. */
size_t DLLENTRY
xio_fread( void* buffer, size_t size, size_t count, XFILE* x )
{
  DEBUGLOG(("xio_fread(%p, %u, %u, %p)\n", buffer, size, count, x));
  int read = count * size;

  ASSERT(x && x->serial == XIO_SERIAL);
  #ifdef NDEBUG
  if( !x || x->serial != XIO_SERIAL ) {
    errno = EBADF;
    return 0;
  }
  #endif

  if (!( x->oflags & XO_READ ))
  { errno = EACCES;
    return 0;
  }
  errno = 0;
  if (read == 0)
    return 0;  
  if (!x->Request())
    return 0;

  size_t rc = 0;
  if (x->protocol->error)
    errno = x->protocol->error;
  else if (!x->protocol->eof)
  { int done = x->protocol->read( buffer, read );
    if( done == read ) {
      rc = count;
    } else if( done >= 0 ) {
      if (done % size == 0 || x->protocol->seek( -( done % size ), XIO_SEEK_CUR ) == -1)
        x->protocol->eof = true;
      rc = done / size;
    } else {
      x->protocol->error = errno;
    }
  }
  x->Release();
  DEBUGLOG(("xio_fread: %u - now at %u\n", rc, x->protocol->tell()));
  return rc;
}

/* Writes up to count items, each of size bytes in length, from buffer
   to the output file. Returns the number of full items successfully
   written, which can be fewer than count if an error occurs. */
size_t DLLENTRY
xio_fwrite( const void* buffer, size_t size, size_t count, XFILE* x )
{
  DEBUGLOG(("xio_fwrite(%p, %u, %u, %p)\n", buffer, size, count, x));
  int write = count * size;

  ASSERT(x && x->serial == XIO_SERIAL);
  #ifdef NDEBUG
  if( !x || x->serial != XIO_SERIAL ) {
    errno = EBADF;
    return 0;
  }
  #endif

  if(!( x->oflags & XO_WRITE )) 
  { errno = EACCES;
    return 0;
  }
  errno = 0;
  if (count == 0)
    return 0;
  if (!x->Request())
    return 0;

  size_t rc = 0;
  if (x->protocol->error)
    errno = x->protocol->error;
  else
  { int done = x->protocol->write( buffer, write );
    if( done == write ) {
      rc = count;
    } else if( done == -1 ) {
      x->protocol->error = errno;
    } else {
      x->protocol->seek( -( done % size ), XIO_SEEK_CUR );
      rc = done / size;
    }
  }
  x->Release();
  DEBUGLOG(("xio_fwrite: %u\n", rc));
  return rc;
}

/* Closes a file pointed to by x. Returns 0 if it successfully closes
   the file, or -1 if any errors were detected. */
int DLLENTRY
xio_fclose( XFILE* x )
{
  DEBUGLOG(("xio_fclose(%p)\n", x));
  ASSERT(x && x->serial == XIO_SERIAL);
  #ifdef NDEBUG
  if( !x || x->serial != XIO_SERIAL ) {
    errno = EBADF;
    return -1;
  }
  #endif

  int ret = x->protocol->close() == 0 ? 0 : -1;
  // cleanup anyway
  delete x;
  DEBUGLOG(("xio_fclose: %i\n", ret));
  return ret;
}

/* Causes an abnormal termination of all current read/write
   operations of the file. All current and subsequent calls can
   raise an error. */
void DLLENTRY
xio_fabort( XFILE* x )
{
  DEBUGLOG(("xio_fabort(%p)\n", x));
  ASSERT(x && x->serial == XIO_SERIAL);
  #ifdef NDEBUG
  if( !x || x->serial != XIO_SERIAL ) {
    errno = EBADF;
    return
  }
  #endif
  x->protocol->close();
}

/* Finds the current position of the file. Returns the current file
   position. On error, returns -1L and sets errno to a nonzero value. */
long DLLENTRY
xio_ftell( XFILE* x )
{
  DEBUGLOG2(("xio_ftell(%p)\n", x));
  ASSERT(x && x->serial == XIO_SERIAL);
  #ifdef NDEBUG
  if( !x || x->serial != XIO_SERIAL ) {
    errno = EBADF;
    return -1;
  }
  #endif
  return x->protocol->tell();
}

/* Changes the current file position to a new location within the file.
   Returns 0 if it successfully moves the pointer. A nonzero return
   value indicates an error. On devices that cannot seek the return
   value is nonzero. */
long DLLENTRY
xio_fseek( XFILE* x, long int offset, int origin )
{
  DEBUGLOG(("xio_fseek(%p, %li, %i)\n", x, offset, origin));
  ASSERT(x && x->serial == XIO_SERIAL);
  #ifdef NDEBUG
  if( !x || x->serial != XIO_SERIAL ) {
    errno = EBADF;
    return -1;
  }
  #endif
  if (!x->Request())
    return -1;
  long ret = x->protocol->seek( offset, origin );
  x->Release();
  DEBUGLOG(("xio_fseek: %li\n", ret));
  return ret;
}

/* Repositions the file pointer associated with stream to the beginning
   of the file. A call to xio_rewind is the same as:
   (void)xio_fseek( x, 0L, XIO_SEEK_SET )
   except that xio_rewind also clears the error indicator for
   the stream. */
void DLLENTRY
xio_rewind( XFILE* x )
{
  xio_fseek( x, 0, XIO_SEEK_SET );
}

/* Returns the size of the file. A return value of -1L indicates an
   error or an unknown size. */
long DLLENTRY
xio_fsize( XFILE* x )
{
  DEBUGLOG(("xio_fsize(%p)\n", x));
  ASSERT(x && x->serial == XIO_SERIAL);
  #ifdef NDEBUG
  if( !x || x->serial != XIO_SERIAL ) {
    errno = EBADF;
    return -1;
  }
  #endif
  long ret = x->protocol->getsize();
  DEBUGLOG(("xio_fsize: %li\n", ret));
  return ret;
}

/* Lengthens or cuts off the file to the length specified by size.
   You must open the file in a mode that permits writing. Adds null
   characters when it lengthens the file. When cuts off the file, it
   erases all data from the end of the shortened file to the end
   of the original file. Returns the value 0 if it successfully
   changes the file size. A return value of -1 shows an error. */
int DLLENTRY
xio_ftruncate( XFILE* x, long size )
{
  DEBUGLOG(("xio_fclose(%p, %lu)\n", x, size));
  ASSERT(x && x->serial == XIO_SERIAL);
  #ifdef NDEBUG
  if( !x || x->serial != XIO_SERIAL ) {
    errno = EBADF;
    return -1;
  }
  #endif
  if(!( x->oflags & XO_WRITE )) 
  { errno = EACCES;
    return -1;
  }
  if (!x->Request())
    return -1;
  int ret = x->protocol->chsize( size );
  x->Release();
  return ret;
}

/* Reads bytes from the current file position up to and including the
   first new-line character (\n), up to the end of the file, or until
   the number of bytes read is equal to n-1, whichever comes first.
   Stores the result in string and adds a null character (\0) to the
   end of the string. The string includes the new-line character, if
   read. If n is equal to 1, the string is empty. Returns a pointer
   to the string buffer if successful. A NULL return value indicates
   an error or an end-of-file condition. */
char* DLLENTRY
xio_fgets( char* string, int n, XFILE* x )
{
  DEBUGLOG(("xio_fgets(%p, %i, %p)\n", string, n, x));

  ASSERT(x && x->serial == XIO_SERIAL);
  #ifdef NDEBUG
  if( !x || x->serial != XIO_SERIAL ) {
    errno = EBADF;
    return NULL;
  }
  #endif
  if (!( x->oflags & XO_READ ))
  { errno = EACCES;
    return NULL;
  }
  if (!x->Request())
    return NULL;
  char* ret = NULL;
  if (x->protocol->error)
    errno = x->protocol->error;
  else if (n <= 1) // Trivial cases
  { if (n <= 0)
    { errno = EINVAL;
    } else
    { *string = 0;
      ret = string;
    }
  } else
  { ret = x->protocol->eof ? NULL : x->protocol->gets(string, n);
  }
  x->Release();
  DEBUGLOG(("xio_fgets: %s, %i\n", ret, x->protocol->error));
  return ret;
}

/* Copies string to the output file at the current position.
   It does not copy the null character (\0) at the end of the string.
   Returns -1 if an error occurs; otherwise, it returns a non-negative
   value. */
int DLLENTRY
xio_fputs( const char* string, XFILE* x )
{
  DEBUGLOG(("xio_fputs(%s, %p)\n", string, x));
  
  ASSERT(x && x->serial == XIO_SERIAL);
  #ifdef NDEBUG
  if( !x || x->serial != XIO_SERIAL ) {
    errno = EBADF;
    return -1;
  }
  #endif
  if(!( x->oflags & XO_WRITE )) 
  { errno = EACCES;
    return -1;
  }
  if (!x->Request())
    return -1;

  int rc;
  if (x->protocol->error)
  { errno = x->protocol->error;
    rc = -1;
  } else
    rc = x->protocol->puts(string);

  x->Release();
  DEBUGLOG(("xio_fputs: %i\n", rc));
  return rc;
}

/* Indicates whether the end-of-file flag is set for the given stream.
   The end-of-file flag is set by several functions to indicate the
   end of the file. The end-of-file flag is cleared by calling xio_rewind,
   xio_fseek, or xio_clearerr for this stream. */
int DLLENTRY
xio_feof( XFILE* x )
{
  DEBUGLOG2(("xio_feof(%p)\n", x));
  ASSERT(x && x->serial == XIO_SERIAL);
  #ifdef NDEBUG
  if( !x || x->serial != XIO_SERIAL ) {
    errno = EBADF;
    return 1;
  }
  #endif
  return x->protocol->eof;
}

/* Tests for an error in reading from or writing to the given stream.
   If an error occurs, the error indicator for the stream remains set
   until you close stream, call xio_rewind, or call xio_clearerr. */
int DLLENTRY
xio_ferror( XFILE* x )
{ 
  DEBUGLOG2(("xio_ferror(%p)\n", x));
  ASSERT(x && x->serial == XIO_SERIAL);
  #ifdef NDEBUG
  if( !x || x->serial != XIO_SERIAL ) {
    errno = EBADF;
    return 1;
  }
  #endif
  return x->protocol->error;
}

/* Resets the error indicator and end-of-file indicator for the
   specified stream. Once set, the indicators for a specified stream
   remain set until your program calls xio_clearerr or xio_rewind.
   xio_fseek also clears the end-of-file indicator. */
void DLLENTRY
xio_clearerr( XFILE* x )
{
  DEBUGLOG(("xio_clearerr(%p)\n", x));
  ASSERT(x && x->serial == XIO_SERIAL);
  #ifdef NDEBUG
  if( !x || x->serial != XIO_SERIAL ) {
    errno = EBADF;
    return;
  }
  #endif
  if (x->Request())
  { x->protocol->error = 0;
    x->protocol->eof   = 0;
    x->Release();
    errno = 0;
  }
}

/* Returns the last error code set by a library call in the current
   thread. Subsequent calls do not reset this error code. */
int DLLENTRY
xio_errno( void ) {
  return errno;
}

/* Maps the error number in errnum to an error message string. */
const char* DLLENTRY
xio_strerror( int errnum )
{
  if( errnum >= CDDBBASEERR ) {
    return XIOcddb::strerror( errnum );
  } else if( errnum >= FTPBASEERR ) {
    return XIOftp::strerror( errnum );
  } else if( errnum >= HTTPBASEERR ) {
    return XIOhttp::strerror( errnum );
  } else if( errnum >= HBASEERR ) {
    return h_strerror( errnum - HBASEERR );
  #ifdef SOCBASEERR
  } else if( errnum >= SOCBASEERR ) {
    return sock_strerror( errnum );
  #endif
  } else {
    return strerror( errnum );
  }
}

/* Sets a handle of a window that are to be notified of changes
   in the state of the library. */
void DLLENTRY
xio_set_observer( XFILE* x, unsigned long window,
                            char* buffer, int buffer_size )
{
  DEBUGLOG(("xio_set_observer(%p, %lu, %p, %i)\n", x, window, buffer, buffer_size));
  ASSERT(x && x->serial == XIO_SERIAL);
  #ifdef NDEBUG
  if( !x || x->serial != XIO_SERIAL ) {
    errno = EBADF;
    return;
  }
  #endif
  if (x->Request())
  { x->s_observer = window;
    x->s_metabuff = buffer;
    x->s_metasize = buffer_size;
    x->Release();
    errno = 0;
  }
}

/* Returns a specified meta information if it is
   provided by associated stream. */
char* DLLENTRY
xio_get_metainfo( XFILE* x, int type, char* result, int size )
{
  DEBUGLOG(("xio_get_metainfo(%p, %i, %p, %i)\n", x, type, result, size));
  ASSERT(x && x->serial == XIO_SERIAL);
  #ifdef NDEBUG
  if( !x || x->serial != XIO_SERIAL ) {
    errno = EBADF;
    return NULL;
  }
  #endif
  if (!size)
  { errno = EINVAL;
    return NULL;
  } else
  { return x->protocol->get_metainfo( type, result, size );
  }
}

/* Returns XIO_NOT_SEEK (0) on streams incapable of seeking,
   XIO_CAN_SEEK (1) on streams capable of seeking and returns
   XIO_CAN_SEEK_FAST (2) on streams capable of fast seeking. */
int DLLENTRY
xio_can_seek( XFILE* x )
{
  DEBUGLOG(("xio_can_seek(%p)\n", x));
  ASSERT(x && x->serial == XIO_SERIAL);
  #ifdef NDEBUG
  if( !x || x->serial != XIO_SERIAL ) {
    errno = EBADF;
    return XIO_NOT_SEEK;
  }
  #endif
  XSFLAGS support = x->protocol->supports();
  DEBUGLOG(("xio_can_seek %x\n", support));
  if( support & XS_CAN_SEEK_FAST )
    return XIO_CAN_SEEK_FAST;
  else if( support & XS_CAN_SEEK )
    return XIO_CAN_SEEK;
  else
    return XIO_NOT_SEEK;
}


/* Returns the read-ahead buffer size. */
int DLLENTRY
xio_buffer_size( void )
{
  return buffer_size;
}

/* Returns fills the buffer before reading state. */
int DLLENTRY
xio_buffer_wait( void )
{
  return buffer_wait;
}

/* Returns value of prefilling of the buffer. */
int DLLENTRY
xio_buffer_fill( void )
{
  return buffer_fill;
}

/* Sets the read-ahead buffer size. */
void DLLENTRY
xio_set_buffer_size( int size )
{
  buffer_size = size;
}

/* Sets fills the buffer before reading state. */
void DLLENTRY
xio_set_buffer_wait( int wait )
{
  buffer_wait = wait;
}

/* Sets value of prefilling of the buffer. */
void DLLENTRY
xio_set_buffer_fill( int percent )
{
  if( percent > 0 && percent <= 100 ) {
    buffer_fill = percent;
  }
}

/* Returns the name of the proxy server. */
void DLLENTRY
xio_http_proxy_host( char* hostname, int size )
{
  Mutex::Lock lock(mutex);
  strlcpy( hostname, http_proxy_host, size );
}

/* Returns the port number of the proxy server. */
int DLLENTRY
xio_http_proxy_port( void )
{
  return http_proxy_port;
}

/* Returns the user name of the proxy server. */
void DLLENTRY
xio_http_proxy_user( char* username, int size )
{
  Mutex::Lock lock(mutex);
  strlcpy( username, http_proxy_user, size );
}

/* Returns the user password of the proxy server. */
void DLLENTRY
xio_http_proxy_pass( char* password, int size )
{
  Mutex::Lock lock(mutex);
  strlcpy( password, http_proxy_pass, size );
}

/* Sets the name of the proxy server. */
void DLLENTRY
xio_set_http_proxy_host( char* hostname )
{
  Mutex::Lock lock(mutex);
  strlcpy( http_proxy_host, hostname, sizeof( http_proxy_host ));
  http_proxy_addr = 0;
}

/* Sets the port number of the proxy server. */
void DLLENTRY
xio_set_http_proxy_port( int port )
{
  http_proxy_port = port;
}

/* Sets the user name of the proxy server. */
void DLLENTRY
xio_set_http_proxy_user( char* username )
{
  Mutex::Lock lock(mutex);
  strlcpy( http_proxy_user, username, sizeof( http_proxy_user ));
}

/* Sets the user password of the proxy server. */
void DLLENTRY
xio_set_http_proxy_pass( char* password )
{
  Mutex::Lock lock(mutex);
  strlcpy( http_proxy_pass, password, sizeof( http_proxy_pass ));
}

/* Returns an internet address of the proxy server.
   Returns 0 if the proxy server is not defined or -1 if
   an error occurs */
unsigned long DLLENTRY
xio_http_proxy_addr( void )
{
  char host[XIO_MAX_HOSTNAME];
  u_long address;

  { Mutex::Lock lock(mutex);
    strlcpy( host, http_proxy_host, sizeof( host ));
    address = http_proxy_addr;
  }

  if( address != 0 && address != (u_long)-1 ) {
    return address;
  }
  if( !*host ) {
    return 0;
  }

  address = so_get_address( host );

  http_proxy_addr = address;
  return address;
}

/* Returns the TCP/IP connection timeout. */
int DLLENTRY
xio_connect_timeout( void )
{
  return socket_timeout;
}

/* Sets the TCP/IP connection timeout. */
void DLLENTRY
xio_set_connect_timeout( int seconds )
{
  socket_timeout = seconds;
}

